from prompt_toolkit.application.current import get_app
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.search import SearchState, SearchDirection
from prompt_toolkit.document import Document

from collections import deque
import logging

logger = logging.getLogger(__name__)


class BetterBuffer(Buffer):
    def __init__(self, *args, search_no_duplicates, **kwargs):
        self.search_no_duplicates = search_no_duplicates

        self._last_working_index = -1

        self._in_search = False
        self._last_search_direction = None
        self._last_search_history = None
        self._search_history = []
        super().__init__(*args, **kwargs)
        original_accept_handler = self.accept_handler

        def _handler(*args, **kwargs):
            self._last_working_index = self.working_index
            return original_accept_handler(*args, **kwargs)

        self.accept_handler = _handler

    def _is_end_of_buffer(self):
        return self.cursor_position == len(self.text)

    def _is_last_history(self):
        return self.working_index == len(self._working_lines) - 1

    def _search_matches(self, i):
        no_duplicates = self.search_no_duplicates
        return not no_duplicates or self._working_lines[i] not in self._search_history

    def _search(
        self,
        search_state: SearchState,
        include_current_position: bool = False,
        count: int = 1):
        # modified to support search_no_duplicates

        assert count > 0

        text = search_state.text
        direction = search_state.direction
        ignore_case = search_state.ignore_case()

        # added by radian
        if direction != self._last_search_direction:
            self._last_search_history = None
            self._search_history = []

        self._in_search = True

        def search_once(
            working_index: int, document: Document
        ):
            """
            Do search one time.
            Return (working_index, document) or `None`
            """
            if direction == SearchDirection.FORWARD:
                # Try find at the current input.
                new_index = document.find(
                    text,
                    include_current_position=include_current_position,
                    ignore_case=ignore_case,
                )

                if new_index is not None:
                    return (
                        working_index,
                        Document(document.text, document.cursor_position + new_index),
                    )
                else:
                    # No match, go forward in the history. (Include len+1 to wrap around.)
                    # (Here we should always include all cursor positions, because
                    # it's a different line.)
                    for i in range(working_index + 1, len(self._working_lines) + 1):
                        i %= len(self._working_lines)

                        # modified by radian
                        if self._search_matches(i):
                            document = Document(self._working_lines[i], 0)
                            new_index = document.find(text, include_current_position=True,
                                                      ignore_case=ignore_case)
                            if new_index is not None:
                                return (i, Document(document.text, new_index))
            else:
                # Try find at the current input.
                new_index = document.find_backwards(text, ignore_case=ignore_case)

                if new_index is not None:
                    return (
                        working_index,
                        Document(document.text, document.cursor_position + new_index),
                    )
                else:
                    # No match, go back in the history. (Include -1 to wrap around.)
                    for i in range(working_index - 1, -2, -1):
                        i %= len(self._working_lines)

                        # modified by radian
                        if self._search_matches(i):
                            document = Document(self._working_lines[i], len(self._working_lines[i]))
                            new_index = document.find_backwards(
                                text, ignore_case=ignore_case)
                            if new_index is not None:
                                return (i, Document(document.text, len(document.text) + new_index))
            return None

        # modified by radian
        working_index = self.working_index
        document = self.document
        for _ in range(count):
            result = search_once(working_index, document)
            if result:
                working_index, document = result

        if result:
            working_index, document = result
            self._last_search_direction = direction
            self._last_search_history = self._working_lines[working_index]
            return (working_index, document.cursor_position)
        else:
            self._last_search_direction = None
            self._last_search_history = None
            self._search_history = []
            return None

    def apply_search(self, *args, **kwargs):
        super().apply_search(*args, **kwargs)
        if self._last_search_history and self._last_search_history not in self._search_history:
            self._search_history.append(self._last_search_history)
        self._in_search = False

    def go_to_next_history(self, i):
        self.go_to_history(i)
        self.history_search_text = ""
        self.history_forward()
        self.cursor_position = len(self.text)

    def auto_up(self, *args, **kwargs):
        if not self.complete_state and not self.selection_state and \
                not self._is_last_history() and self._is_end_of_buffer():
            self.history_backward()
            self.cursor_position = len(self.text)
        else:
            super().auto_up(*args, **kwargs)

    def auto_down(self, *args, **kwargs):
        if not self.complete_state and not self.selection_state and \
                not self._is_last_history() and self._is_end_of_buffer():
            self.history_forward()
            self.cursor_position = len(self.text)
        elif not self.complete_state and not self.selection_state and \
                self._is_last_history() and len(self.text) == 0 and self._last_working_index >= 0 \
                and self._last_working_index < len(self._working_lines) - 1:
            # down arrow after commiting a history line
            self.go_to_next_history(self._last_working_index)
            self._last_working_index = -1
        else:
            super().auto_down(*args, **kwargs)

    def _reset_searching(self):
        self._in_search = False
        self._last_search_direction = None
        self._last_search_history = None
        self._search_history = []

    def reset(self, *args, **kwargs):
        self._reset_searching()
        super().reset(*args, **kwargs)


class ModalBuffer(BetterBuffer):
    def __init__(self, *args, session, **kwargs):
        self.session = session
        super().__init__(*args, **kwargs)
        # we don't use load_history_if_not_yet_loaded because it breaks ctrl-o
        # https://github.com/prompt-toolkit/python-prompt-toolkit
        self._reset_history()

    # def _set_working_mode(self):
    #     if not self._in_search and self._is_last_history():
    #         self.working_mode = self.session.current_mode

    # def _change_working_mode(self, index):
    #     if index < len(self._working_lines_mode) - 1:
    #         mode = self._working_lines_mode[index]
    #         self.session.activate_mode(mode)
    #     elif self.working_mode:
    #         self.session.activate_mode(self.working_mode)

    def _history_mode_matches(self, i):
        if i == len(self._working_lines) - 1:
            return True
        else:
            spec = self.session.current_mode_spec
            next_mode = self._working_lines_mode[i]
            if next_mode == spec.name:
                return True
            elif next_mode in self.session.specs and \
                    spec.history_book == self.session.specs[next_mode].history_book:
                return True
        return False

    def _history_matches(self, i):
        return super()._history_matches(i) and self._history_mode_matches(i)

    def load_history_if_not_yet_loaded(self):
        # use _reset_history instead
        pass

    def _search_matches(self, i):
        return super()._search_matches(i) and self._history_mode_matches(i)

    def _search(self, *args, **kwargs):
        # self._set_working_mode()
        res = super()._search(*args, **kwargs)
        # if res:
        #     self._change_working_mode(res[0])
        return res

    def auto_up(self, *args, **kwargs):
        # self._set_working_mode()
        super().auto_up(*args, **kwargs)
        # self._change_working_mode(self.working_index)

    def auto_down(self, *args, **kwargs):
        # self._set_working_mode()
        super().auto_down(*args, **kwargs)
        # self._change_working_mode(self.working_index)

    def append_to_history(self) -> None:
        if not self.session.add_history:
            return
        if not self.session.current_mode_spec.keep_history:
            return
        if self.text:
            history_strings = self.history.get_strings()
            history_modes = self.history.get_modes()
            if not len(history_strings) or \
                    history_strings[-1] != self.text or \
                    history_modes[-1] != self.session.current_mode:
                self.history.append_string(self.text, self.session.current_mode)

    def _reset_history(self):
        self._working_lines_mode = deque([None])
        self.working_mode = None
        for m, item in self.history.load():
            self._working_lines.appendleft(item)
            self._working_lines_mode.appendleft(m)
            self._Buffer__working_index += 1

    def reset(self, *args, **kwargs):
        super().reset(*args, **kwargs)
        self._reset_history()



import os
import sys
import subprocess


def main(cleanup=None):
    import optparse
    import os
    import sys
    import subprocess
    from rchitect.utils import Rhome, rversion
    from radian import __version__

    try:
        # failed to import jedi on demand in some edge cases.
        import jedi  # noqa
    except ImportError:
        pass

    parser = optparse.OptionParser("usage: radian")
    parser.add_option("-v", "--version", action="store_true", dest="version", help="Get version")
    parser.add_option("--r-binary", dest="r", help="Path to R binary")
    parser.add_option("--profile", dest="profile", help="Path to .radian_profile, ignore both global and local profiles")
    parser.add_option("-q", "--quiet", "--silent", action="store_true", dest="quiet", help="Don't print startup message")
    parser.add_option("--no-environ", action="store_true", dest="no_environ", help="Don't read the site and user environment files")
    parser.add_option("--no-site-file", action="store_true", dest="no_site_file", help="Don't read the site-wide Rprofile")
    parser.add_option("--no-init-file", action="store_true", dest="no_init_file", help="Don't read the user R profile")
    parser.add_option("--local-history", action="store_true", dest="local_history", help="Force using local history file")
    parser.add_option("--global-history", action="store_true", dest="global_history", help="Force using global history file")
    parser.add_option("--no-history", action="store_true", dest="no_history", help="Don't load any history files")
    parser.add_option("--vanilla", action="store_true", dest="vanilla", help="Combine --no-history --no-environ --no-site-file --no-init-file")
    parser.add_option("--save", action="store_true", dest="save", help="Do save workspace at the end of the session")
    parser.add_option("--ask-save", action="store_true", dest="ask_save", help="Ask to save R data")
    parser.add_option("--restore-data", action="store_true", dest="restore_data", help="Restore previously saved objects")
    parser.add_option("--debug", action="store_true", dest="debug", help="Debug mode")
    parser.add_option("--coverage", action="store_true", dest="coverage", help=optparse.SUPPRESS_HELP)
    parser.add_option("--cprofile", action="store_true", dest="cprofile", help=optparse.SUPPRESS_HELP)

    # we accept these options, but never check them
    parser.add_option("--no-save", action="store_true", dest="no_save", help=optparse.SUPPRESS_HELP)
    parser.add_option("--no-restore-data", action="store_true", dest="no_restore_data", help=optparse.SUPPRESS_HELP)
    parser.add_option("--no-restore-history", action="store_true", dest="no_restore_history", help=optparse.SUPPRESS_HELP)
    parser.add_option("--no-restore", action="store_true", dest="no_restore", help=optparse.SUPPRESS_HELP)
    parser.add_option("--no-readline", action="store_true", dest="no_readline", help=optparse.SUPPRESS_HELP)
    parser.add_option("--interactive", action="store_true", dest="interactive", help=optparse.SUPPRESS_HELP)

    options, args = parser.parse_args()

    if options.r:
        os.environ["R_BINARY"] = options.r

    r_home = Rhome()

    if options.version:
        if r_home:
            r_binary = os.path.normpath(os.path.join(r_home, "bin", "R"))
            r_version = rversion(r_home)
        else:
            r_binary = "NA"
            r_version = "NA"
        print("radian version: {}".format(__version__))
        print("r executable: {}".format(r_binary))
        print("r version: {}".format(r_version))
        print("python executable: {}".format(sys.executable))
        print("python version: {:d}.{:d}.{:d}".format(
            sys.version_info.major,
            sys.version_info.minor,
            sys.version_info.micro))
        return

    os.environ["RADIAN_VERSION"] = __version__
    os.environ["RADIAN_COMMAND_ARGS"] = " ".join(
        ["--" + k.replace("_", "-") for k, v in options.__dict__.items() if v])

    if not r_home:
        raise RuntimeError("Cannot find R binary. Expose it via the `PATH` variable.")

    if not sys.platform.startswith("win"):
        libPath = os.path.join(r_home, "lib")
        ldpaths = os.path.join(r_home, "etc", "ldpaths")
        if (
            "R_LD_LIBRARY_PATH" not in os.environ
            or libPath not in os.environ["R_LD_LIBRARY_PATH"]
        ):
            if os.path.exists(ldpaths):
                R_LD_LIBRARY_PATH = (
                    subprocess.check_output(
                        '. "{}"; echo $R_LD_LIBRARY_PATH'.format(ldpaths),
                        shell=True,
                    )
                    .decode("utf-8")
                    .strip()
                )
            elif "R_LD_LIBRARY_PATH" in os.environ:
                R_LD_LIBRARY_PATH = os.environ["R_LD_LIBRARY_PATH"]
            else:
                R_LD_LIBRARY_PATH = libPath
            if libPath not in R_LD_LIBRARY_PATH:
                R_LD_LIBRARY_PATH = "{}:{}".format(libPath, R_LD_LIBRARY_PATH)
            os.environ["R_LD_LIBRARY_PATH"] = R_LD_LIBRARY_PATH
            # respect R_ARCH variable?
            if sys.platform == "darwin":
                ld_library_var = "DYLD_FALLBACK_LIBRARY_PATH"
            else:
                ld_library_var = "LD_LIBRARY_PATH"
            if ld_library_var in os.environ:
                LD_LIBRARY_PATH = "{}:{}".format(
                    R_LD_LIBRARY_PATH, os.environ[ld_library_var]
                )
            else:
                LD_LIBRARY_PATH = R_LD_LIBRARY_PATH
            os.environ[ld_library_var] = LD_LIBRARY_PATH
            if sys.argv[0].endswith("radian"):
                os.execv(sys.argv[0], sys.argv)
            else:
                os.execv(
                    sys.executable,
                    [sys.executable, "-m", "radian"] + sys.argv[1:],
                )

    RadianApplication(r_home, ver=__version__).run(options, cleanup=cleanup)


def get_app():
    return RadianApplication.instance


class RadianApplication():
    instance = None
    r_home = None

    def __init__(self, r_home, ver):
        RadianApplication.instance = self
        self.r_home = r_home
        self.ver = ver
        super(RadianApplication, self).__init__()

    def set_env_vars(self, options):
        if options.vanilla:
            options.no_history = True
            options.no_environ = True
            options.no_site_file = True
            options.no_init_file = True

        if options.no_environ:
            os.environ["R_ENVIRON"] = ""
            os.environ["R_ENVIRON_USER"] = ""

        if options.no_site_file:
            os.environ["R_PROFILE"] = ""

        if options.no_init_file:
            os.environ["R_PROFILE_USER"] = ""

        if options.local_history:
            if not os.path.exists(".radian_history"):
                open(".radian_history", 'w+').close()

        doc_dir = os.path.join(self.r_home, "doc")
        include_dir = os.path.join(self.r_home, "include")
        share_dir = os.path.join(self.r_home, "share")
        if not (os.path.isdir(doc_dir) and os.path.isdir(include_dir) and os.path.isdir(share_dir)):
            try:
                paths = subprocess.check_output([
                    os.path.join(self.r_home, "bin", "R"), "--slave", "--vanilla", "-e",
                    "cat(paste(R.home('doc'), R.home('include'), R.home('share'), sep=':'))"
                ])
                doc_dir, include_dir, share_dir = paths.decode().split(":")
            except Exception:
                pass

        os.environ["R_DOC_DIR"] = doc_dir
        os.environ["R_INCLUDE_DIR"] = include_dir
        os.environ["R_SHARE_DIR"] = share_dir

        # enable crayon on windows
        # we use CMDER_ROOT as a temporary workaround
        if sys.platform.startswith("win"):
            if "CMDER_ROOT" not in os.environ:
                os.environ["CMDER_ROOT"] = "NA"

    def run(self, options, cleanup=None):
        from .prompt_session import create_radian_prompt_session
        from .console import create_read_console, create_write_console_ex
        import rchitect
        from . import dispatch  # noqa
        from . import rutils, settings

        self.set_env_vars(options)

        args = ["radian", "--quiet", "--no-restore-history"]

        if sys.platform != "win32":
            args.append("--no-readline")

        if options.no_environ:
            args.append("--no-environ")

        if options.no_site_file:
            args.append("--no-site-file")

        if options.no_init_file:
            args.append("--no-init-file")

        if options.save:
            args.append("--save")
        elif options.ask_save is not True:
            args.append("--no-save")

        if options.restore_data is not True:
            args.append("--no-restore-data")

        # disable the code injection of rchitect to reticulate::py_discover_config
        os.environ["RCHITECT_RETICULATE_CONFIG"] = "0"
        # enable signal handlers
        os.environ["RCHITECT_REGISTER_SIGNAL_HANDLERS"] = "1"

        rchitect.init(args=args, register_signal_handlers=True)

        if sys.platform.startswith("win"):
            rutils.set_utf8()

        try:
            rutils.source_radian_profile(options.profile)
        except RuntimeError as e:
            print("Got an error while loading radian profile")
            print(e)

        settings = settings.radian_settings
        settings.load()
        self.session = create_radian_prompt_session(options, settings)

        rchitect.def_callback(name="read_console")(create_read_console(self.session))
        rchitect.def_callback(name="write_console_ex")(
            create_write_console_ex(self.session, settings.stderr_format))

        rutils.load_custom_key_bindings()

        if cleanup:
            rutils.register_cleanup(cleanup)

        from . import reticulate
        reticulate.configure()

        # run user on load hooks
        try:
            rutils.run_on_load_hooks()
        except Exception as e:
            print("Error in running user hooks")
            print(e)

        # print welcome message
        if options.quiet is not True:
            self.session.app.output.write(rchitect.interface.greeting())

        rchitect.loop()



import signal
import re
from contextlib import contextmanager

from .settings import radian_settings as settings
from .rutils import is_long_non_ascii_multiline
from rchitect import console

TERMINAL_CURSOR_AT_BEGINNING = [True]

SUPPRESS_STDOUT = False
SUPPRESS_STDERR = False
ANSI_ESCAPE_RE = re.compile(r'\x1B\[[0-?]*[ -/]*[@-~]')


def normalize(string):
    return ANSI_ESCAPE_RE.sub('', string.replace('\r\n', '\n'))


@contextmanager
def suppress_stderr(suppress=True):
    """
    It is used in completion to avoid running prompt-toolkit nestedly
    """
    global SUPPRESS_STDERR
    OLD_SUPPRESS_STDERR = SUPPRESS_STDERR
    SUPPRESS_STDERR = suppress
    try:
        yield
    finally:
        console.flush()
        SUPPRESS_STDERR = OLD_SUPPRESS_STDERR


def sigint_handler(signum, frame):
    raise KeyboardInterrupt()


def ask_input(s):
    orig_handler = signal.getsignal(signal.SIGINT)
    # allow Ctrl+C to throw KeyboardInterrupt in callback
    signal.signal(signal.SIGINT, sigint_handler)
    try:
        return input(s)
    finally:
        signal.signal(signal.SIGINT, orig_handler)


def native_prompt(app, message):
    # c.f. run_coroutine_in_terminal of prompt_toolkit
    with suppress_stderr(False):
        console.flush()
        app.output.flush()
        app._running_in_terminal = True
        try:
            with app.input.detach():
                with app.input.cooked_mode():
                    return ask_input(message)
        except KeyboardInterrupt:
            app.output.write_raw("\n")
            raise
        finally:
            app._running_in_terminal = False
            app.renderer.reset()
            app._request_absolute_cursor_position()
            app._redraw()


def create_read_console(session):
    interrupted = [False]

    def _read_console(message, add_history=1):
        app = session.app

        if app.is_running:
            # fallback to `input` if `read_console` is called nestedly
            return native_prompt(app, message)

        session._prompt_message = message
        current_mode_spec = session.current_mode_spec

        if interrupted[0]:
            interrupted[0] = False
            if not session.current_mode_spec.sticky_on_sigint:
                session.activate_mode(session.mode_to_be_activated())
            if current_mode_spec.insert_new_line_on_sigint:
                app.output.write_raw("\n")
        elif not TERMINAL_CURSOR_AT_BEGINNING[0] or \
                (settings.insert_new_line and current_mode_spec.insert_new_line):
            app.output.write_raw("\n")

        text = None

        while text is None:
            if not session.current_mode_spec.sticky:
                session.activate_mode(session.mode_to_be_activated())

            try:
                text = session.prompt(add_history=add_history)

            except KeyboardInterrupt:
                interrupted[0] = True
                raise

            except Exception as e:
                if isinstance(e, EOFError):
                    # todo: confirmation in "r" mode
                    return None
                else:
                    print("unexpected error was caught.")
                    print("please report to https://github.com/randy3k/radian for such error.")
                    print(e)
                    import traceback
                    traceback.print_exc()
                    import os
                    os._exit(1)

            if text is None and settings.insert_new_line and current_mode_spec.insert_new_line:
                app.output.write_raw("\n")

        return text

    _text_stored = ["", 0, False]  # text, startpos, sent_by_line

    def read_console(message, add_history):
        if session.current_mode in ["r", "browse"]:
            # this code is needed to allow new line breaks with strings, see #377
            if _text_stored[0]:
                text = _text_stored[0][_text_stored[1]:]
            else:
                text = _read_console(message, add_history)
                if text and is_long_non_ascii_multiline(text):
                    # make sure the text is evaluated at once
                    # in case it includes long multiline strings, see #379
                    text = "{\n" + text + "\n}"
                    _text_stored[2] = True

                _text_stored[0] = text
                _text_stored[1] = 0

            if text:
                sent_by_line = _text_stored[2]
                index = text.find('\n')
                if sent_by_line and index >= 0:
                    _text_stored[1] += index + 1
                    text = text[:index]
                else:
                    _text_stored[0] = ""
                    _text_stored[1] = 0
                    _text_stored[1] = False
        else:
            text = _read_console(message, add_history)

        return text

    return read_console


def create_write_console_ex(session, stderr_format):
    app = session.app
    output = app.output
    from prompt_toolkit.utils import is_windows

    write_console_ex = None

    if is_windows():
        from prompt_toolkit.output.win32 import Win32Output
        if isinstance(output, Win32Output):
            # we use print_formatted_text to support ANSI sequences in older Windows
            from prompt_toolkit.formatted_text import ANSI
            from prompt_toolkit.shortcuts import print_formatted_text

            def write_console_ex(buf, otype):
                if otype == 0:
                    if not SUPPRESS_STDOUT:
                        buf = buf.replace("\r\n", "\n")
                        sbuf = buf.split("\r")
                        for i, b in enumerate(sbuf):
                            print_formatted_text(ANSI(b), end="", output=output)
                            if i < len(sbuf) - 1:
                                output.write("\r")
                        output.flush()
                        buf = normalize(buf)
                        if buf:
                            TERMINAL_CURSOR_AT_BEGINNING[0] = buf.endswith("\n")
                else:
                    if not SUPPRESS_STDERR:
                        buf = buf.replace("\r\n", "\n")
                        sbuf = buf.split("\r")
                        for i, b in enumerate(sbuf):
                            print_formatted_text(
                                ANSI(stderr_format.format(b)), end="", output=output)
                            if i < len(sbuf) - 1:
                                output.write("\r")
                        output.flush()
                        buf = normalize(buf)
                        if buf:
                            TERMINAL_CURSOR_AT_BEGINNING[0] = buf.endswith("\n")

    if not write_console_ex:
        def write_console_ex(buf, otype):
            if otype == 0:
                if not SUPPRESS_STDOUT:
                    output.write_raw(buf)
                    output.flush()
                    buf = normalize(buf)
                    if buf:
                        TERMINAL_CURSOR_AT_BEGINNING[0] = buf.endswith("\n")
            else:
                if not SUPPRESS_STDERR:
                    output.write_raw(stderr_format.format(buf))
                    output.flush()
                    buf = normalize(buf)
                    if buf:
                        TERMINAL_CURSOR_AT_BEGINNING[0] = buf.endswith("\n")

    return write_console_ex



import datetime
import os
from prompt_toolkit.history import InMemoryHistory, FileHistory


class ModelHistory:
    def load(self):
        if not self._loaded:
            self._loaded_strings = list(self.load_history_strings())
            self._loaded = True

        for item in self._loaded_strings:
            yield item

    def append_string(self, string: str, mode) -> None:
        self._loaded_strings.insert(0, (mode, string))
        self.store_string(string, mode)

    def get_strings(self):
        return [s for _, s in reversed(self._loaded_strings)]

    def get_modes(self):
        return [m for m, _ in reversed(self._loaded_strings)]


class ModalInMemoryHistory(ModelHistory, InMemoryHistory):
    def store_string(self, string, mode):
        pass


class ModalFileHistory(ModelHistory, FileHistory):
    def __init__(self, history_file, max_history_size):
        self.max_history_size = max_history_size
        super().__init__(history_file)

    def load_history_strings(self):
        strings = []
        lines = []
        mode = [None]
        breaks = []

        def add() -> None:
            if lines:
                # Join and drop trailing newline.
                string = (mode[0], "".join(lines)[:-1])
                strings.append(string)

        if os.path.exists(self.filename):
            with open(self.filename, "rb") as f:
                for i, line_bytes in enumerate(f):
                    line = line_bytes.decode("utf-8", errors="replace")

                    if line.startswith('# mode: '):
                        mode[0] = line.replace('# mode: ', '').strip()
                    elif line.startswith("+"):
                        lines.append(line[1:])
                    else:
                        add()
                        if lines:
                            breaks.append(i)
                        lines = []

                add()

            if len(breaks) > max(self.max_history_size, 10):
                # trim history if it is too big
                with open(self.filename, "r+", encoding="utf-8") as f:
                    backup = f.readlines()
                    f.seek(0)
                    f.truncate()
                    trimed = backup[breaks[-round(self.max_history_size * 0.9)]+1:]
                    f.writelines(trimed)

        # Reverse the order, because newest items have to go first.
        return reversed(strings)

    def store_string(self, string, mode):
        # Save to file.
        with open(self.filename, 'ab') as f:
            def write(t):
                f.write(t.encode('utf-8'))

            write('\n# time: %s UTC' % datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S"))
            write('\n# mode: %s\n' % mode)
            for line in string.split('\n'):
                write('+%s\n' % line)



from .buffer import ModalBuffer

from prompt_toolkit import PromptSession
from prompt_toolkit.application import Application
from prompt_toolkit.application.current import get_app
from prompt_toolkit.auto_suggest import DynamicAutoSuggest
from prompt_toolkit.completion import DynamicCompleter, ThreadedCompleter
from prompt_toolkit.enums import DEFAULT_BUFFER
from prompt_toolkit.filters import Condition, emacs_mode
from prompt_toolkit.key_binding.key_bindings import \
    KeyBindings, DynamicKeyBindings, merge_key_bindings
from prompt_toolkit.validation import DynamicValidator
from prompt_toolkit.shortcuts.prompt import is_true, CompleteStyle
from prompt_toolkit.utils import to_str

from collections import OrderedDict
from typing import cast


# TODO: allow lines from different modes when replying history

class ModeSpec():
    def __init__(
            self,
            name,
            on_activated=None,
            on_dectivated=None,
            keep_history=True,
            history_book=None,
            prompt_key_bindings=None,
            **kwargs):
        self.name = name
        self.on_activated = on_activated
        self.on_dectivated = on_dectivated
        self.keep_history = keep_history
        if history_book:
            self.history_book = history_book
        else:
            self.history_book = name
        self.prompt_key_bindings = prompt_key_bindings
        for key in kwargs:
            if key not in PromptSession._fields:
                raise KeyError("unknown field", key)
            setattr(self, key, kwargs[key])


class ModalPromptSession(PromptSession):
    _spec_class = ModeSpec
    _current_mode = None
    _default_settings = {}
    _specs = OrderedDict()
    _inputhook = None

    # new settings
    add_history = True
    search_no_duplicates = False

    def _check_args(self, kwargs):
        if "specs" in kwargs:
            specs = kwargs["specs"]
            for m in specs.values():
                assert isinstance(m, ModeSpec)

    def _filter_args(self, kwargs):
        for key in ["add_history", "search_no_duplicates"]:
            if key in kwargs:
                setattr(self, key, kwargs[key])
                del kwargs[key]

    def __init__(self, inputhook=None, *args, **kwargs):
        self._inputhook = inputhook
        self._check_args(kwargs)
        self._filter_args(kwargs)
        super().__init__(*args, **kwargs)
        self._backup_settings()

    # for backward compatibility
    @property
    def modes(self):
        return self._specs

    @property
    def specs(self):
        return self._specs

    @property
    def current_mode(self):
        return self._current_mode

    @property
    def current_mode_spec(self):
        return self.specs[self.current_mode]

    def register_mode(self, name, **kwargs):
        spec = self._spec_class(name, **kwargs)
        self.specs[spec.name] = spec
        if len(self.specs) == 1:
            self.activate_mode(spec.name)
        else:
            self.activate_mode(self.current_mode, force=True)

    def unregister_mode(self, spec_or_name):
        if isinstance(spec_or_name, str):
            del self.specs[spec_or_name]
        else:
            del self.specs[next(iter(k for k, v in self.specs.items() if v == spec_or_name))]

    def activate_mode(self, name, force=False):
        if name not in self.specs:
            raise Exception("no such mode")

        spec = self.specs[name]

        if self.current_mode == spec.name and not force:
            return

        if self.current_mode:
            current_spec = self.specs[self.current_mode]
            if current_spec.on_dectivated:
                current_spec.on_dectivated(self)

        self._current_mode = spec.name

        self._restore_settings()
        for name in self._fields:
            if name != "key_bindings":
                if hasattr(spec, name):
                    setattr(self, name, getattr(spec, name))

        self.key_bindings = merge_key_bindings(
            [DynamicKeyBindings(lambda: self.specs[self.current_mode].prompt_key_bindings)] +
            [
                m.key_bindings for m in self.specs.values()
                if hasattr(m, "key_bindings") and m.key_bindings
            ]
        )

        if spec.on_activated:
            spec.on_activated(self)

    def _backup_settings(self):
        for name in self._fields:
            self._default_settings[name] = getattr(self, name)

    def _restore_settings(self):
        for name in self._fields:
            setattr(self, name, self._default_settings[name])

    def _create_default_buffer(self):
        """
        radian modifications
            supports both complete_while_typing and enable_history_search

        Create and return the default input buffer.
        """
        dyncond = self._dyncond

        # Create buffers list.
        def accept(buff) -> bool:
            """Accept the content of the default buffer. This is called when
            the validation succeeds."""
            cast(Application[str], get_app()).exit(result=buff.document.text)
            return True  # Keep text, we call 'reset' later on.

        return ModalBuffer(
            name=DEFAULT_BUFFER,
            # Make sure that complete_while_typing is disabled when
            # enable_history_search is enabled. (First convert to Filter,
            # to avoid doing bitwise operations on bool objects.)
            complete_while_typing=Condition(
                lambda: is_true(self.complete_while_typing)
                # and not is_true(self.enable_history_search)
                and not self.complete_style == CompleteStyle.READLINE_LIKE
            ),
            validate_while_typing=dyncond("validate_while_typing"),
            enable_history_search=dyncond("enable_history_search"),
            validator=DynamicValidator(lambda: self.validator),
            completer=DynamicCompleter(
                lambda: ThreadedCompleter(self.completer)
                if self.complete_in_thread and self.completer
                else self.completer
            ),
            history=self.history,
            auto_suggest=DynamicAutoSuggest(lambda: self.auto_suggest),
            accept_handler=accept,
            tempfile_suffix=lambda: to_str(self.tempfile_suffix or ""),
            tempfile=lambda: to_str(self.tempfile or ""),
            session=self,
            search_no_duplicates=self.search_no_duplicates
        )

    def _create_application(self, *args, **kwargs):
        app = super()._create_application(*args, **kwargs)

        kb = KeyBindings()

        # operate-and-get-next
        @kb.add('c-o', filter=emacs_mode)
        def _(event):
            buff = event.current_buffer
            working_index = buff.working_index
            buff.validate_and_handle()

            def set_working_index() -> None:
                buff.go_to_next_history(working_index)

            event.app.pre_run_callables.append(set_working_index)

        app._default_bindings = merge_key_bindings([app._default_bindings, kb])

        return app

    def prompt(self, *args, **kwargs):
        self._check_args(kwargs)
        self._filter_args(kwargs)
        if args:
            raise Exception("positional arguments are deprecated")

        backup = self._default_settings.copy()
        for name in self._fields:
            if name in kwargs:
                value = kwargs[name]
                if value is not None:
                    setattr(self._default_settings, name, value)

        orig_mode = self.current_mode
        try:
            result = super().prompt(inputhook=self._inputhook, **kwargs)
        except KeyboardInterrupt:
            self._default_settings = backup.copy()
            self.activate_mode(orig_mode, force=True)
            raise KeyboardInterrupt
        finally:
            self._default_settings = backup.copy()

        # prompt will restore settings, we need to reactivate current mode
        self.activate_mode(self.current_mode, force=True)
        return result



import os
import re
import sys
import time

from .lineedit.prompt import ModalPromptSession, ModeSpec
from .lineedit.history import ModalInMemoryHistory, ModalFileHistory
from prompt_toolkit.enums import EditingMode
from prompt_toolkit.formatted_text import ANSI
from prompt_toolkit.layout.processors import HighlightMatchingBracketProcessor
from prompt_toolkit.lexers import PygmentsLexer
from prompt_toolkit.styles import style_from_pygments_cls
from prompt_toolkit.utils import is_windows, get_term_environment_variable
from prompt_toolkit.validation import Validator
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory

from pygments.styles import get_style_by_name

from rchitect import rcopy, rcall, robject
from rchitect.interface import roption, setoption, parse_text_complete, \
    process_events, peek_event, polled_events

from . import shell
from .key_bindings import create_r_key_bindings, create_shell_key_bindings, create_key_bindings
from .completion import RCompleter, SmartPathCompleter
from .io import CustomInput, CustomOutput
from .lexer import CustomSLexer as SLexer


PROMPT = "\x1b[34mr$>\x1b[0m "
SHELL_PROMPT = "\x1b[31m#!>\x1b[0m "
BROWSE_PROMPT = "\x1b[33mBrowse[{}]>\x1b[0m "
BROWSE_PATTERN = re.compile(r"Browse\[([0-9]+)\]> $")
VI_MODE_PROMPT = "\x1b[34m[{}]\x1b[0m "


class RadianModeSpec(ModeSpec):
    def __init__(
            self,
            name,
            prompt_message=None,
            is_activated=None,
            callback=None,
            sticky=False,
            sticky_on_sigint=False,
            insert_new_line=False,
            insert_new_line_on_sigint=False,
            **kwargs):
        self.prompt_message = prompt_message
        self.is_activated = is_activated
        self.callback = callback
        self.sticky = sticky
        self.sticky_on_sigint = sticky_on_sigint
        self.insert_new_line = insert_new_line
        self.insert_new_line_on_sigint = insert_new_line_on_sigint
        super().__init__(name, **kwargs)


class RadianPromptSession(ModalPromptSession):
    _spec_class = RadianModeSpec
    _prompt_message = ""

    def mode_to_be_activated(self):
        for name in reversed(self.specs):
            spec = self.specs[name]
            if spec.is_activated and spec.is_activated(self):
                return name
        return "unknown"

    def prompt(self, *args, **kwargs):
        text = super().prompt(*args, **kwargs)
        current_mode_spec = self.current_mode_spec
        if current_mode_spec.callback:
            text = current_mode_spec.callback(self)

        return text


def apply_settings(session, settings):
    setoption("prompt", settings.prompt)

    if settings.auto_width:
        output_width = session.app.output.get_size().columns
        if output_width:
            setoption("width", output_width)

    # necessary on windows
    setoption("menu.graphics", False)

    def askpass(message):
        app = session.app
        if app.is_running:
            from getpass import getpass
            return getpass(message)
        else:
            from prompt_toolkit import prompt
            return prompt(message, is_password=True)

    if not roption("askpass"):
        setoption("askpass", robject(askpass, convert=True))

    # enables completion of installed package names
    if rcopy(rcall(("utils", "rc.settings"), "ipck")) is None:
        rcall(("utils", "rc.settings"), ipck=True)


def create_radian_prompt_session(options, settings):

    local_history_file = settings.local_history_file
    global_history_file = settings.global_history_file

    if options.no_history:
        history = ModalInMemoryHistory()
    elif not options.global_history and os.path.exists(local_history_file):
        history = ModalFileHistory(os.path.abspath(local_history_file), settings.history_size)
    else:
        history_file = os.path.join(os.path.expanduser(global_history_file))
        history_file = os.path.expandvars(history_file)
        history_file_dir = os.path.dirname(history_file)
        if not os.path.exists(history_file_dir):
            os.makedirs(history_file_dir, 0o700)
        history = ModalFileHistory(history_file, settings.history_size)

    if is_windows():
        output = None
    else:
        output = CustomOutput.from_pty(sys.stdout, term=get_term_environment_variable())

    def vi_mode_prompt():
        if session.editing_mode == EditingMode.VI and settings.show_vi_mode_prompt:
            im = session.app.vi_state.input_mode.value
            vi_mode_prompt = settings.vi_mode_prompt
            if isinstance(vi_mode_prompt, str):
                return vi_mode_prompt.format(str(im)[3:6])
            else:
                return vi_mode_prompt[str(im)[3:6]]
        return ""

    def message():
        if session.current_mode_spec.prompt_message:
            return ANSI(
                vi_mode_prompt() + session.current_mode_spec.prompt_message(session._prompt_message)
                )
        else:
            return session._prompt_message

    if settings.editing_mode in ["vim", "vi"]:
        editing_mode = EditingMode.VI
    else:
        editing_mode = EditingMode.EMACS

    def get_inputhook():
        # make testing more robust
        if "RADIAN_NO_INPUTHOOK" in os.environ:
            return None

        terminal_width = [None]

        def _(context):
            output_width = session.app.output.get_size().columns
            if output_width and terminal_width[0] != output_width:
                terminal_width[0] = output_width
                setoption("width", max(terminal_width[0], 20))

            while True:
                if context.input_is_ready():
                    break
                try:
                    if peek_event():
                        with session.app.input.detach():
                            with session.app.input.rare_mode():
                                process_events()
                    else:
                        polled_events()

                except Exception:
                    pass
                time.sleep(1.0 / 30)

        return _

    session = RadianPromptSession(
        message=message,
        style=style_from_pygments_cls(get_style_by_name(settings.color_scheme)),
        editing_mode=editing_mode,
        history=history,
        enable_history_search=True,
        search_no_duplicates=settings.history_search_no_duplicates,
        search_ignore_case=settings.history_search_ignore_case,
        enable_suspend=True,
        input=CustomInput(sys.stdin),
        output=output,
        auto_suggest=AutoSuggestFromHistory() if settings.auto_suggest else None,
        inputhook=get_inputhook()
    )

    input_processors = []
    if settings.highlight_matching_bracket:
        input_processors.append(HighlightMatchingBracketProcessor())

    r_key_bindings = create_r_key_bindings(parse_text_complete)

    session.register_mode(
        name="r",
        prompt_message=lambda x: x,
        is_activated=lambda session: session._prompt_message == settings.prompt,
        history_book="r",
        insert_new_line=True,
        multiline=settings.indent_lines,
        completer=RCompleter(timeout=settings.completion_timeout),
        complete_while_typing=settings.complete_while_typing,
        lexer=PygmentsLexer(SLexer),
        tempfile_suffix=".R",
        input_processors=input_processors,
        key_bindings=create_key_bindings(),
        prompt_key_bindings=r_key_bindings
    )

    browse_level = [""]

    def browse_activator(session):
        message = session._prompt_message
        if BROWSE_PATTERN.match(message):
            browse_level[0] = BROWSE_PATTERN.match(message).group(1)
            return True
        else:
            return False

    class BrowseValidator(Validator):
        """
        As a pre-accept processor.
        """
        def validate(self, document):
            text = document.text
            if settings.history_ignore_browser_commands:
                if text.strip() in ["n", "s", "f", "c", "cont", "Q", "where", "help"]:
                    session.add_history = False

    session.register_mode(
        name="browse",
        is_activated=browse_activator,
        prompt_message=lambda _: settings.browse_prompt.format(browse_level[0]),
        history_book="r",
        insert_new_line=True,
        multiline=settings.indent_lines,
        completer=RCompleter(timeout=settings.completion_timeout),
        complete_while_typing=settings.complete_while_typing,
        validator=BrowseValidator(),
        lexer=PygmentsLexer(SLexer),
        tempfile_suffix=".R",
        input_processors=input_processors,
        prompt_key_bindings=r_key_bindings
    )

    def shell_process_text(session):
        text = session.default_buffer.text
        if text.strip():
            shell.run_command(text)

    session.register_mode(
        name="shell",
        prompt_message=lambda _: settings.shell_prompt,
        callback=shell_process_text,
        sticky=True,
        sticky_on_sigint=False,
        insert_new_line=True,
        multiline=settings.indent_lines,
        completer=SmartPathCompleter(),
        complete_while_typing=settings.complete_while_typing,
        lexer=None,
        input_processors=input_processors,
        prompt_key_bindings=create_shell_key_bindings()
    )

    session.register_mode(
        "unknown",
        prompt_message=lambda x: x,
        insert_new_line=False,
        complete_while_typing=False,
        keep_history=False,
        lexer=None,
        completer=None,
        prompt_key_bindings=None,
        input_processors=[]
    )

    apply_settings(session, settings)

    return session
